# -*- coding: utf-8 -*-
"""Take normalized raw data find the peaks in it."""

import argparse
import sys
import warnings
from os.path import abspath

import numpy as np
from matplotlib import pyplot as plt
from scipy.stats import norm

from rendseq.file_funcs import make_new_dir, open_wig, write_wig


def _populate_trans_mat(z_scores, peak_center, spread, trans_m, states):
    """Calculate the Vertibi Algorithm transition matrix.

    Parameters
    ----------
        -z_scores (2xn array): - required: first column is position (ie bp
            location) second column is a modified z_score for that position.
        -peak_center (float): the mean of the emission probability distribution
            for the peak state.
        -spread (float): the standard deviation of the peak emmission
            distribution.
        -trans_m (matrix): the transition probabilities between states.
        -states (matrix): how internal and peak are represented in the wig file

    Raises
    ------
        -ValueError if the HMM parameters are unable to find any likely paths.
            This is at risk of happening if the emission probabilities for
            either the peak state or the internal state look very different from
            the true distribution, or the parameters selected only allow a very
            unlikely path - for example if p_to_p = 1 and you try to force the
            system to stay in the peak state.
    """
    print("Calculating Transition Matrix")
    trans_1 = np.zeros([len(states), len(z_scores)])
    trans_2 = np.zeros([len(states), len(z_scores)]).astype(int)
    trans_1[:, 0] = 1

    # emission probabilities matrix (2 x n)
    probs = np.zeros(shape=(len(states), len(z_scores)))
    probs[0, :] = norm.pdf(
        z_scores[:, 1]
    )  # note zscores[:,1] is where all the zscores are
    probs[1, :] = norm.pdf(z_scores[:, 1], peak_center, spread)

    # Vertibi Algorithm:
    for i in range(1, len(z_scores)):
        # we use log probabilities for computational reasons. -Inf means 0 probability
        paths = np.zeros((len(states), len(states)))
        paths += np.expand_dims(
            trans_1[:, i - 1], axis=1
        )  # adds log(trans_1_sub) column-wise
        paths += np.log(trans_m)  # adds log(trans_m) element-wise
        paths += np.expand_dims(
            np.log(probs[:, i]), axis=0
        )  # adds log(probs_sub) row-wise
        trans_2[:, i] = np.argmax(paths, axis=0)
        trans_1[:, i] = np.max(paths, axis=0)

    if np.any(np.isinf(trans_1[:, -1])):
        raise ValueError(
            "".join(
                [
                    "Parameters Provided are too far from the true",
                    "distribution for a valid path to be found.  Please ",
                    "change your paramters for hmm_peaks and try again.",
                ]
            )
        )
    return trans_1, trans_2


def hmm_peaks(z_scores, i_to_p=1 / 1000, p_to_p=1 / 1.5, peak_center=10, spread=2):
    """Fit peaks to the provided z_scores data set using the vertibi algorithm.

    Parameters
    ----------
        -z_scores (2xn array): - required: first column is position (ie bp
            location) second column is a modified z_score for that position.
        -i_to_p (float): value should be between zero and 1, represents
            probability of transitioning from inernal state to peak state. The
            default value is 1/2000, based on asseumption of geometrically
            distributed transcript lengths with mean length 1000. Should be a
            robust parameter.
        -p_to_p (float): The probability of a peak to peak transition.  Default
            1/1.5.
        -peak_center (float): the mean of the emission probability distribution
            for the peak state.
        -spread (float): the standard deviation of the peak emmission
            distribution.

    Returns
    -------
        -peaks: a 2xn array with the first column being position and the second
            column being a peak assignment.
    """
    print("Finding Peaks")
    max_z = peak_center + spread
    # cap the z scores at peak_center + spread to eliminate very large zscores.
    trim_zscores = z_scores[:, :]
    trim_zscores[:, 1] = np.where(z_scores[:, 1] > max_z, max_z, z_scores[:, 1])
    trans_m = np.asarray(
        [[(1 - i_to_p), (i_to_p)], [p_to_p, (1 - p_to_p)]]
    )  # transition probability
    states = [1, 100]  # how internal and peak are represented in the wig file
    trans_1, trans_2 = _populate_trans_mat(
        trim_zscores, peak_center, spread, trans_m, states
    )
    peaks = np.zeros([len(z_scores), 2])
    peaks[:, 0] = trim_zscores[:, 0]
    # Now we trace backwards and find the most likely path:
    max_inds = np.zeros([len(peaks)]).astype(int)
    max_inds[-1] = int(np.argmax(trans_1[:, len(trans_1)]))
    peaks[-1, 1] = states[max_inds[-1]]
    for index in reversed(list(range(1, len(peaks)))):
        max_inds[index - 1] = trans_2[max_inds[index], index]
        peaks[index - 1, 1] = states[max_inds[index - 1]]
    print(f"Found {sum(peaks[:,1] == states[1])} Peaks")
    return peaks


def _make_kink_fig(save_file, seen, exp, pnts, thresh):
    """Create a figure comparing the obs vs exp z score distributions.

    Parameters
    ----------
        - save_fig (str) - the name of the file to save the plot to.
        - seen (1xn array) - the obs number of positions with a given z score or greater
        - exp (1xn array) - the exp number of positions with a given z score or greater
        - pnts (1xn array) - the z score values/x axis of the plot.
        - thresh (int) - the threshold value which was ultimately selected.
    """
    plt.plot(pnts, seen, label="Observed")
    plt.plot(pnts, exp, label="Expected")
    plt.plot([thresh, thresh], [max(exp) * 10, min(exp) / 10], label="Threshold")
    plt.yscale("log")
    plt.ylabel("Number of Positions with Z score Greater than or equal to")
    plt.xlabel("Z score")
    plt.legend()
    plt.savefig(save_file)


def _calc_thresh(z_scores, method, kink_img="./kink.png"):
    """Calculate a threshold for z-scores file using the method provided.

    Parameters
    ----------
        - z_scores (2xn array): the calculated z scores, where the first column
            represents the nt position and the second represents a z score.
        - method (string): the name of the threshold calculating method to use.

    Returns
    -------
        - threshold (float): the calculated threshold.
    """
    methods = ["expected_val", "kink"]
    thresh = 15
    if method == "expected_val":  # threshold such that num peaks exp < 1.
        p_val = 1 / len(z_scores)  # note this method is dependent on genome size
        thresh = round(norm.ppf(1 - p_val), 1)
    elif method == "kink":  # where the num z_scores exceeds exp num by 10000x
        factor_exceed = 10000
        pnts = np.arange(0, 20, 0.1)
        seen = [0 for i in range(len(pnts))]
        exp = [0 for i in range(len(pnts))]
        thresh = -1
        for ind, point in enumerate(pnts):
            seen[ind] = np.sum(z_scores[:, 1] > point)
            exp[ind] = (1 - norm.cdf(point)) * len(z_scores)
            if seen[ind] >= factor_exceed * exp[ind] and thresh == -1:
                thresh = point

        _make_kink_fig(kink_img, seen, exp, pnts, thresh)

    else:
        warnings.warn(
            "\n".join(
                [
                    f"The method selected ({method}) is not supported.",
                    f"Please select one from {methods}.",
                    f"Defaulting to threshold of {thresh}.",
                ]
            )
        )
    return thresh


def thresh_peaks(z_scores, thresh=None, method="kink"):
    """Find peaks by calling z-scores above a threshold as a peak.

    Parameters
    ----------
        - z_scores - a 2xn array of nt positions and zscores at that pos.
        - thresh - the threshold value to use.  If none is provided it will be
            automatically calculated.
        - method - the method to use to automatically calculate the z score
            if none is provided.  Default method is "kink"
    """
    if thresh is None:
        thresh = _calc_thresh(z_scores, method)
    peaks = np.zeros([len(z_scores), 2])
    peaks[:, 0] = z_scores[:, 0]
    peaks[:, 1] = (z_scores[:, 1] > thresh).astype(int)
    return peaks


def parse_args_make_peaks(args):
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Can run from the\
                                        commmand line.  Please pass a \
                                        zscore file and select a method \
                                        for peak fitting."
    )
    parser.add_argument("filename", help="Location of the zscore file")
    parser.add_argument(
        "method",
        help='User must pass the desired peak\
                                        fitting method.  Choose "thesh" \
                                        or "hmm"',
    )
    parser.add_argument(
        "--save_file",
        help="Save the z_scores file as a new\
                                        wig file in addition to returning the\
                                        z_scores.  Default = True",
        default=True,
    )
    return parser.parse_args(args)


def main_make_peaks():
    """Run the main peak making from command line."""
    args = parse_args_make_peaks(sys.argv[1:])
    filename = args.filename
    z_scores, chrom = open_wig(filename)
    if args.method == "thresh":
        print(f"Using the thresholding method to find peaks for {filename}")
        peaks = thresh_peaks(z_scores)
    elif args.method == "hmm":
        print(f"Using the hmm method to find peaks for {filename}")
        peaks = hmm_peaks(z_scores)
    else:
        raise ValueError("{args.method} is not a valid peak finding method, see --help")
    if args.save_file:
        filename = abspath(filename).replace("\\", "/")
        file_loc = filename[: filename.rfind("/")]
        peak_dir = make_new_dir([file_loc, "/Peaks/"])
        file_start = filename[filename.rfind("/") + 1 : filename.rfind(".wig")]
        peak_file = "".join([peak_dir, file_start, "_peaks.wig"])
        write_wig(peaks, peak_file, chrom)
        print(f"Wrote peaks to {peak_file}")


if __name__ == "__main__":
    main_make_peaks()



# -*- coding: utf-8 -*-
"""Provide general utility functions used in various rendseq analyses."""

import numpy as np


def find_peak_locs(peaks, new_data):
    """
    Find the indices where there are peaks.

    Parameters
    ----------
        - peaks: a 2xn array with the first column being position and the second
            column being a peak assignment.

    Returns
    -------
        - indices: an array with all the indices where peaks can be found.
    """
    peak_value = max(peaks[:, 1])
    peak_locs = peaks[list(np.where(peaks[:, 1] == peak_value)[0]), 0]
    indices = [0 for i in range(len(peak_locs))]
    for p_ind, bp_val in enumerate(peak_locs):
        indices[p_ind] = np.where(new_data[:, 0] == bp_val)[0][0]
    return indices


def zero_padding(data):
    """
    Zero-pad the reads array.

    Add zero values to genomic positions missing from the reads data because
    there were no transcript ends mapped to that position.

    Parameters
    ----------
        - data: a 2xn array containing raw rendseq reads

    Returns
    -------
        - zero_padded_wig: a 2xm array containing zero-padded reads
    """
    print("Zero-padding the reads...")
    ind_start = int(data[0, 0])
    ind_end = int(data[-1, 0])
    indexes = [i for i in range(ind_start, ind_end + 1)]
    values = [0 for i in range(len(indexes))]
    cur_ind = 0
    for ind, bp in enumerate(indexes):
        while data[cur_ind, 0] < bp:
            cur_ind += 1
        if data[cur_ind, 0] == bp:
            values[ind] = data[cur_ind, 1]
        else:
            values[ind] = 0
    zero_padded_wig = np.zeros([len(indexes), 2])
    zero_padded_wig[:, 0] = indexes
    zero_padded_wig[:, 1] = values
    print("Done zero-padding.")
    return zero_padded_wig


def smooth_peaks(peaks_arr, raw_arr):
    """
    Create a "smoothed" or shaved version of the peaks array.

    Take all the peaks called via the peak calling script and
    replace them with the mean of reads after it.

    Parameters
    ----------
        - peaks_arr: a 2xn array with the first column being position and the second
            column being a peak assignment.
        - raw_arr: a 2xn array containing raw rendseq reads

    Returns
    -------
        - smoothed: a 2xn array containing the shaved peaks array
    """
    print("Shaving peaks...")
    smoothed = raw_arr.copy()
    indices = find_peak_locs(peaks_arr, smoothed)
    for p_ind in indices:
        gap = 5
        mean = np.mean(smoothed[p_ind - 3 * gap : p_ind - gap, 1])
        std = np.std(raw_arr[p_ind - 3 * gap : p_ind - gap, 1])
        smoothed[p_ind - gap : p_ind + gap + 1, 1] = abs(
            np.random.normal(mean, std, 2 * gap + 1)
        )
    print("Done shaving peaks.")
    return smoothed



# -*- coding: utf-8 -*-
"""
Find ramps and steps in Rendseq data.

Classes:

    ReadsData
    StretchData

Functions:
    arr_inrange(np.array, np.array, np.array) -> np.array
    output_to_wig(StretchData, StretchData, string, string)
"""

from datetime import datetime

import numpy as np
import pandas as pd
import scipy.stats as stats

import rendseq.file_funcs as ff


class ReadsData:
    """
    A class to represent reads data with further processing.

    Provides methods to compute cumulative sum, blurred slope,
    and the derivative of the blurred slope.

    Attributes
    ----------
        - reads: a 2xn array of zero-padded read data
        - dat_idx: a 2xn array of indices, from 0 to len(reads)
        - csum: a length n array of the cumulative sum
        - blurred_data: a length n array of the blurred data
            i.e. the blurred slope of the cumulative sum
        - blurred_deriv: a length n array of the derivative after blurring
            i.e. the derivative of the blurred data


    Methods
    -------
        - process_data(window_width = 400, blur_width = 100):
            Calculates a cumulative sum, blurred data, and derivative.

        - __blurred_slope__(idx, window_width, blur_width):
            Helper method for process_data.

        - __find_window_edges__(center, half_window, blur_width, edge):
            Helper method for blurred_slope.
    """

    def __init__(self, reads):
        """Construct a ReadsData object using a 2nx array of reads."""
        self.reads = reads
        self.dat_idx = np.arange(len(self.reads))

    def __blurred_slope__(self, idx, window_width, blur_width):
        """
        Perform the blurred slope calculation when vectorized.

        Helper method for process_data.
        """
        if idx % 500000 == 0:
            print("index: ", idx, ", time: ", datetime.now())

        # find edges
        half_window = int(1 / 2 * window_width)
        half_blur = int(1 / 2 * blur_width)
        left = self.__find_window_edges__(idx, half_window, half_blur, edge="left")
        right = self.__find_window_edges__(idx, half_window, half_blur, edge="right")

        # if edge indices are invalid, return np.nan as blurry slope
        if left < 0:  # if left is smaller than first possible index
            return np.nan
        if right > len(self.csum) - 1:  # if right is larger than last possible index
            return np.nan

        # take the slope
        arr = self.csum[
            left:right
        ]  # subset of array representing the cumulative reads to take a slope over
        upper_mean = arr[-(2 * half_blur) :].mean()  # the first 2*blur_width reads
        lower_mean = arr[: (2 * half_blur)].mean()  # the last 2*blur_width reads
        slope = (upper_mean - lower_mean) / (2 * half_window)

        return slope

    def __find_window_edges__(self, center, half_window, half_blur, edge):
        """
        Find the left or right edge of the window to take the blured slope over.

        Helper method for blurred_slope.
        """
        if edge == "left":
            return int(center - half_window - half_blur)
        if edge == "right":
            return int(center + half_window + half_blur)
        else:
            raise ValueError("edge must be 'left' or 'right' ")

    def process_data(self, window_width=400, blur_width=200):
        """
        Calculate a cumulative sum, blurred data, and derivative.

        Updates the corresponding attributes.
        """
        # cumulative sum
        self.csum = np.cumsum(self.reads[:, 1])

        # blurred slope of cumulative sum
        print("Calculating blurred slope... This can take a few minutes")
        vblurred_slope = np.vectorize(self.__blurred_slope__)
        self.blurred_data = vblurred_slope(
            self.dat_idx, window_width=window_width, blur_width=blur_width
        )
        print("Done calculating blurred slope")

        # derivative of blurred slope
        self.blurred_deriv = np.gradient(self.blurred_data, self.reads[:, 0])


class StretchData:
    """
    A class to represent stretches of positive or negative slope.

    Provides methods to find contiguous stretches of increasing or
    decreasing read density, filter them based on length, and output results

    Attributes
    ----------
        - reads: a 2xn array of zero-padded read data
        - deriv: a length n array of the derivative after blurring
            i.e. the derivative of the blurred data
        - kind: str, either 'positive' or 'negative'
            Refelects whether stretches of positive slope (increasing)
            or negative slope (decreasing) are being detected
        - search_idx: array of indices of `deriv` where deriv shifts signs
        - search_starts: array of indices of `search_idx` where stretches of
            positive/negative (depending on `kind`) values of deriv begin
        - search_sizes: array of lengths of positive/negative stretches
        - starts: array of indices of `search_idx` where sufficiently long
            (>min_gap) stretches start
        - sizes: array of lengths of sufficiently long positive/negative stretches
        - start_datidx: array of indices in `deriv` where sufficiently long
           stretches start
        - start_genomidx: array of genomic positions where sufficiently long
           stretches start
        - end_genomidx: array of genomic positions where sufficiently long stretches end

    Methods
    -------
        - stretch_search():
            Finds all contiguous stretches of positive or negative derivative,
            depending on `kind`.
        - filter_strenth_length(min_gap):
            Filters out stretches that are smaller than the `min_gap` cutoff.
        - ttest_stretches(gap_len = 50, window_len = 100, pval_cutoff = 1):
            Applies a t-test to the read density before and after each stretch.
        - output_csv(filename, sort_length = True):
            Creates a pandas dataframe of sufficiently long stretches, with genomic
            position and stretch length. Sorts by stretch length (from long to short)
            by default. Outputs the dataframe as a .csv file.

    """

    def __init__(self, readsobj, kind):
        """
        Construct a StretchData object.

        Parameters
        ----------
        readsobj: a ReadsData object with a defined `blurred_deriv` attribute.
        kind: either "positive" (increasing stretch) or "negative" (decreasing stretch)
        """
        self.reads = readsobj.reads
        self.deriv = readsobj.blurred_deriv
        self.kind = kind

    def stretch_search(self):
        """
        Find all contiguous stretches of positive or negative derivative.

        Finds increasing stretches (blurred_deriv > 0) with `kind` = "positive".
        Finds decreasing stretches (blurred_deriv < 0) with `kind` = "negative".
        Note: considers values that are exactly 0 (rare) as both
            negative and positive.
        After finding all stretches, updates search_idx, search_starts,
            and search_sizes.
        """
        # Sizes of gaps between contiguous positive stretches = length of
        #     negative stretches
        kind = self.kind
        if kind == "positive":
            idx = np.where(self.deriv <= 0)[0]
        elif kind == "negative":
            idx = np.where(self.deriv >= 0)[0]
        else:
            raise ValueError("kind must be 'positive' or 'negative'")

        # Find discontinuity in indices
        idx_shift = idx[1:]  # indices shifted by one
        aligned = np.stack([idx[:-1], idx_shift])  # stack the two arrays
        stretch_starts = np.where(aligned[0] + 1 != aligned[1])[
            0
        ]  # list of indices (indices of aligned) where a gap begins
        stretch_sizes = aligned[1, stretch_starts] - aligned[0, stretch_starts]

        self.search_idx = idx
        self.search_starts = stretch_starts
        self.search_sizes = stretch_sizes

    def filter_stretch_length(self, min_gap):
        """
        Keep stretches that are longer than the `min_gap` argument.

        Determines the genomic locations of the sufficiently long stretches.
        Updates starts, sizes, start_datidx, start_genomidx, end_genomidx.
        """
        # filter features by length
        passing = np.where(self.search_sizes > min_gap)
        start_passing = self.search_starts[passing]
        sizes_passing = self.search_sizes[passing]

        # store indices of features
        # `datidx`: array of indices of the data (0 to n);
        #      n = length of the reads, cumulative sum, self.deriv, arrays
        # `genomidx`: array of actual genomic positions.
        start_datidx = self.search_idx[start_passing]  # indices of
        start_genomidx = self.reads[start_datidx, 0]  # genomic positions

        self.starts = start_passing
        self.sizes = sizes_passing
        self.start_datidx = start_datidx
        self.start_genomidx = start_genomidx
        self.end_genomidx = self.start_genomidx + self.sizes

    def ttest_stretches(self, gap_len=50, window_len=100, pval_cutoff=1):
        """
        Apply a t-test to the read density before and after each stretch.

        By default, this function is not used and the default p-value cutoff is set to 1
        to avoid filtering out any stretches on the basis of this test.
        """

        def ttest_candidate_shift(
            dat, start_datidx, stretch_len, gap_len, window_len, **kwargs
        ):
            """
            T-test an individual stretch of increasing/decreasing data.

            For a given stretch, look left and right (at some gap) of the strech
            and t-test the left vs. right to see if the increase/decrease
            is persistent and significant.
            """
            # set indices defining the left window of values and
            #     right window of values to test against each other
            left_start = start_datidx - gap_len - window_len
            left_end = start_datidx - gap_len
            right_start = start_datidx + stretch_len + gap_len
            right_end = start_datidx + stretch_len + gap_len + window_len

            # account for stretches very beginning and end of data
            if left_start < 0:
                left_start = 0
            if left_end < 0:
                left_end = 0
            if right_start > len(dat) - 1:
                right_start = len(dat) - 1
            if right_end > len(dat) - 1:
                right_end = len(dat) - 1

            left = dat[left_start:left_end, 1]
            right = dat[right_start:right_end, 1]

            t, p = stats.ttest_ind(
                left, right, equal_var=False, nan_policy="omit", **kwargs
            )

            return p

        # alternative hypothesis type
        if self.kind == "positive":
            alt = "less"
        elif self.kind == "negative":
            alt = "greater"
        else:
            raise ValueError("kind must be 'positive' or 'negative'")

        # apply ttest to all putative stretches
        results = []
        for idx, length in zip(self.start_datidx, self.sizes):
            p = ttest_candidate_shift(
                self.reads, idx, length, gap_len=50, window_len=100, alternative=alt
            )
            results.append(p)
        pvals = np.array(results)

        # filter based on p-value
        condition = pvals < pval_cutoff
        start = self.start_genomidx[condition]
        end = self.start_genomidx[condition] + self.sizes[condition]

        self.pvals = pvals
        self.start_genomidx = start
        self.end_genomidx = end

    def output_csv(self, filename, sort_length=True):
        """
        Create a Pandas dataframe of sufficiently long stretches.

        Output includes genomic position and stretch length. Sorts by stretch length
        (from long to short) by default. Outputs the dataframe as a .csv file.
        """
        # create the output dataframe
        compiled = np.stack([self.start_genomidx, self.sizes])
        compiled_df = pd.DataFrame(
            compiled.transpose(), columns=["genomic location", "length"]
        )

        # sort from longest to shortest features if desired
        if sort_length:
            compiled_df = compiled_df.sort_values("length", ascending=False)

        # output to csv
        compiled_df.to_csv(filename, index=False)


def arr_inrange(idx, feature_starts, feature_ends):
    """
    Locate indices within idx that lie between feature_starts and feature_ends.

    Parameters
    ----------
        - idx: an array of length p of position values
        - feature_starts: an array of length l of feature starting positions
        - feature_ends: an array of length l of feature ending positions

    Returns
    -------
        - inrange: a boolean array of length p
    """

    def isin_range(idx, feature_starts, feature_ends):
        """
        Check if a value lies between two corresponding bookends.

        Helper function for arr_inrange that gets vectorized.
        In the non-vectorized form, checks if a value (idx) lies between any
            two corresponding values of `feature_starts` and `feature_ends`
        """
        greater = np.where(feature_starts > idx)[
            0
        ]  # all instances of being greater than query
        if len(greater) == 0:
            return False
        first_greater = greater[0]
        start = feature_starts[first_greater - 1]
        end = feature_ends[first_greater - 1]
        if (idx > start) & (idx < end):
            return True
        else:
            return False

    # vectorize function and return vectorized result
    v_isin_range = np.vectorize(isin_range, excluded=["feature_starts", "feature_ends"])
    inrange = v_isin_range(
        idx=idx, feature_starts=feature_starts, feature_ends=feature_ends
    )
    return inrange


def output_to_wig(inc_stretchobj, dec_stretchobj, wig_output_filename, chrom):
    """
    Write a .wig file containing putative ramps and steps.

    Increasing stretches labeled with +1 and decreasing stretches are labeled with -1.

    Parameters
    ----------
        - inc_stretchobj: a StretchData object with increasing stretches
            (kind = 'positive')
        - dec_sretchobj: a StretchData object with decreasing stretches
            (kind = 'negative')
        - wig_output_filename: name of the output file
        - chrom: the appropriate chrom for the output .wig
    """
    # find which genomic indices lie between the starting and ending
    #     genomic indices for each feature
    dec_inrange = arr_inrange(
        idx=dec_stretchobj.reads[:, 0],
        feature_starts=dec_stretchobj.start_genomidx,
        feature_ends=dec_stretchobj.end_genomidx,
    )
    inc_inrange = arr_inrange(
        idx=inc_stretchobj.reads[:, 0],
        feature_starts=inc_stretchobj.start_genomidx,
        feature_ends=inc_stretchobj.end_genomidx,
    )

    # write wig: with increasing stretches as +1 and decreasing stretches as -1
    output_wig = dec_stretchobj.reads.copy()
    output_wig[:, 1] = 0
    output_wig[dec_inrange, 1] = -1
    output_wig[inc_inrange, 1] = 1
    print("Writing the ramps/steps file to a .wig")
    ff.write_wig(output_wig, wig_output_filename, chrom)
    print("Done writing the wig.")



# -*- coding: utf-8 -*-
"""Functions for fetching, creating, and opening raw and processed data files."""

from os import mkdir
from os.path import isdir

from numpy import asarray, delete, where
from pandas import read_csv


def _validate_reads(reads):
    """Make sure the given reads meet our format requirements.

    Parameters
    ----------
        -reads (2xn array): a 2xn array with the first column being position
            and the second column being the count at that position (raw read,
            z_score etc)

    Returns
    -------
    NoneType

    Raises
    ------
        Throws exceptions if reads are not correctly formatted
    """
    # Check if reads are empty
    if len(reads) < 1:
        raise ValueError("requires non-empty reads")

    # If reads.shape doesn't work, then it's not an np array
    try:
        shape = reads.shape
    except Exception:
        raise ValueError(f"reads must be numpy array, not {type(reads)}")

    # Must have two columns
    if shape[1] != 2:
        raise ValueError(f"reads must be (n,2), not {reads.shape}")


def write_wig(wig_track, wig_file_name, chrom_name):
    """Write provided data to the wig file.

    Parameters
    ----------
        - wig_track (required) - the wig data you wish to write (in 2xn array)
        - wig_file_name (string) - the new file you will write to
    """
    _validate_reads(wig_track)
    d_inds = where(wig_track[:, 0] < 1)
    wig_track = delete(wig_track, d_inds, axis=0)
    with open(wig_file_name, "w+", encoding="utf-8") as wig_file:
        wig_file.write("track type=wiggle_0\n")
        wig_file.write(f"variableStep chrom={chrom_name}\n")
        for i in range(len(wig_track)):
            wig_file.write(f"{int(wig_track[i,0])}\t{wig_track[i,1]}\n")


def open_wig(filename):
    """Open the provided wig file and return the contents into a 2xn array.

    Parameters
    ----------
        -filename (string) - required: the string containing the location of
            the filename you desire to open!

    Returns
    -------
        -reads (2xn array): a 2xn array with the first column being position
            and the second column being the count at that position (raw read,
            z_score etc)
    """
    # first we will read the chrom from the second line in the wig file:
    with open(filename, "r", encoding="utf8") as file:
        try:
            next(file)
        except StopIteration:
            raise ValueError(f"{filename} appears to have zero lines")

        line = file.readline()
        chrom = line[line.rfind("=") + 1 :].rstrip()
    # next we read all the wig file data and return that if it's valid:
    reads = asarray(read_csv(filename, sep="\t", header=1, names=["bp", "count"]))
    _validate_reads(reads)
    return reads, chrom


def make_new_dir(dir_parts):
    """Create a new directory and return valid path to it.

    Parameters
    ----------
        - dir_parts  - a list of strings to be joined to make the directory name

    Returns
    -------
        - dir_str - the directory name
    """
    dir_str = "".join(dir_parts)
    if not isdir(dir_str):
        mkdir(dir_str)
    return dir_str



# -*- coding: utf-8 -*-
"""set package information for rendseq pip package."""
from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="rendseq",
    version="0.1.8",
    description="Package for RendSeq Data Analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/miraep8/rendseq",
    author="Mirae Parker",
    author_email="rendseq@mit.edu",
    license="MIT",
    packages=["rendseq"],
    python_requires=">=3.6",
)



# -*- coding: utf-8 -*-
"""Functions needed for z-score transforming raw rendSeq data."""
import argparse
import sys
import warnings
from os.path import abspath

import numpy as np
from numpy import mean, nan, std, zeros

from rendseq.file_funcs import _validate_reads, make_new_dir, open_wig, write_wig


def _get_lower_reads(cur_ind, target_start, target_stop, reads):
    """Calculate the padded lower reads index in range for the z-score calculation."""
    cur_ind = max(min(cur_ind, len(reads) - 1), 0)
    vals = np.random.normal(0, 1, [abs(target_stop - target_start) + 1, 1])
    ind = 0
    while cur_ind >= 0 and reads[cur_ind, 0] >= target_stop:
        vals[ind] = reads[cur_ind, 1]
        cur_ind -= 1
        ind += 1
    return vals


def _get_upper_reads(cur_ind, target_start, target_stop, reads):
    """Fetch the padded upper reads needed for z score calculation with zero padding."""
    cur_ind = min(max(cur_ind, 0), len(reads) - 1)

    vals = np.random.normal(0, 1, [abs(target_stop - target_start) + 1, 1])
    ind = 0
    while reads[cur_ind, 0] < target_stop and not cur_ind >= len(reads) - 1:
        vals[ind] = reads[cur_ind, 1]
        ind += 1
        cur_ind += 1
    return vals


def _calc_z_score(vals, calc_val):
    """Calculate a z-score given a value, mean, and standard deviation.

    vals = values to calculate the z score with respect to.
    calc_val = value to calculate z score for.

    NOTE: The z_score() of a constant vector is 0
    """
    v_std = std(vals)
    v_mean = mean(vals)
    if nan in [v_mean, v_std]:
        return calc_val
    if v_std == 0:
        return 0 if v_mean == calc_val else (calc_val - v_mean) / 0.2
    return (calc_val - v_mean) / v_std


def _remove_outliers(vals, method="remove_by_std"):
    """Normalize window of reads by removing outliers (values 2.5 std > mean).

    Parameters
    ----------
        -vals: an array of raw read values to be processed

    Returns
    -------
        -new_v: another array of raw values which has had the extreme values
            removed.
    """
    normalized_vals = vals

    if method == "remove_by_std" and len(vals) > 1:
        v_std = std(vals)
        if v_std != 0:
            normalized_vals = [v for v in vals if abs(_calc_z_score(vals, v)) < 2.5]

    return normalized_vals


def _validate_gap_window(gap, w_sz):
    """Check that gap and window size are reasonable in r/l_score_helper."""
    if w_sz < 1:
        raise ValueError("Window size must be larger than 1 to find a z-score")
    if gap < 0:
        raise ValueError("Gap size must be at least zero to find a z-score")
    if gap == 0:
        warnings.warn(
            "Warning...a gap size of 0 includes the current position.", stacklevel=2
        )


def z_scores(reads, gap=5, w_sz=50):
    """Perform modified z-score transformation of reads.

    Parameters
    ----------
        -reads 2xn array - raw rendseq reads
        -gap (interger):   number of reads surround the current read of
            interest that should be excluded in the z_score calculation.
        -w_sz (integer): the max distance (in nt) away from the current position
            one should include in zscore calulcation.

    Returns
    -------
        -z_scores (2xn array): a 2xn array with the first column being position
            and the second column being the z_score.
    """
    _validate_gap_window(gap, w_sz)
    _validate_reads(reads)
    # make array of zscores - same length as raw reads, trimming based on window size:
    z_scores = zeros([len(reads) - 2 * (gap + w_sz), 2])

    # first column of return array is the location of the raw reads
    z_scores[:, 0] = reads[gap + w_sz : len(reads) - (gap + w_sz), 0]

    # Iterate through each valid read, recording z-score
    for i in range((gap + w_sz + 1), (len(reads) - (gap + w_sz))):
        # calculate the z score with values from the left:
        i_score_pos = i - (gap + w_sz)
        if reads[i, 1] > 5:
            l_vals = _get_upper_reads(
                i + gap, reads[i, 0] + gap, reads[i, 0] + gap + w_sz, reads
            )
            l_score = _calc_z_score(_remove_outliers(l_vals), reads[i, 1])

            # calculate z score with reads from the right:
            r_vals = _get_lower_reads(
                i - gap, reads[i, 0] - gap, reads[i, 0] - gap - w_sz, reads
            )
            r_score = _calc_z_score(_remove_outliers(r_vals), reads[i, 1])

            # set the zscore to be the smaller valid score of the left/right scores.
            z_scores[i_score_pos, 1] = (
                r_score if abs(r_score) < abs(l_score) else l_score
            )
        else:
            z_scores[i_score_pos, 1] = 1

    return z_scores


def parse_args_zscores(args):
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Takes raw read file and\
                                        makes a modified z-score for each\
                                        position. Takes several optional\
                                        arguments"
    )
    parser.add_argument(
        "filename",
        help="Location of the raw_reads file that\
                                        will be processed using this function.\
                                        Should be a properly formatted wig\
                                        file.",
    )
    parser.add_argument(
        "--gap",
        help="gap (interger):   number of reads\
                                        surround the current read of interest\
                                        that should be excluded in the z_score\
                                        calculation. Defaults to 5.",
        default=5,
    )
    parser.add_argument(
        "--w_sz",
        help="w_sz (integer): the max dis (in nt)\
                                        away from the current position one\
                                        should include in zscore calulcation.\
                                        Default to 50.",
        default=50,
    )
    parser.add_argument(
        "--save_file",
        help="Save the z_scores file as a new\
                                        wig file in addition to returning the\
                                        z_scores.  Default = True",
        default=True,
    )
    return parser.parse_args(args)


def main_zscores():
    """Run Z-score calculations.

    Effect: Writes messages to standard out. If --save-file flag,
    also writes output to disk.
    """
    args = parse_args_zscores(sys.argv[1:])

    # Calculate z-scores
    filename = args.filename
    print(f"Calculating zscores for file {filename}.")
    reads, chrom = open_wig(filename)
    z_score = z_scores(reads, gap=int(args.gap), w_sz=int(args.w_sz))

    # Save file, if applicable
    if args.save_file:
        _save_zscore(filename, z_score, chrom)
    print(
        "\n".join(
            [
                "Ran zscores.py with the following settings:",
                f"gap: {args.gap}, w_sz: {args.w_sz},",
                f"file_name: {args.filename}",
            ]
        )
    )


def _save_zscore(filename, z_score, chrom):
    filename = abspath(filename).replace("\\", "/")
    file_loc = filename[: filename.rfind("/")]
    z_score_dir = make_new_dir([file_loc, "/Z_scores"])
    file_start = filename[filename.rfind("/") : filename.rfind(".")]
    z_score_file = "".join([z_score_dir, file_start, "_zscores.wig"])
    write_wig(z_score, z_score_file, chrom)
    print(f"Wrote z_scores to {z_score_file}")


if __name__ == "__main__":
    main_zscores()
